import pytest

from vakt.parser import compile_regex
from vakt.exceptions import InvalidPatternError


@pytest.mark.parametrize('phrase, start, end, output', [
    ('foo-bar-<.*>', '<', '>', r'^foo\-bar\-(.*)$'),
    ('foo-bar-[.*]', '[', ']', r'^foo\-bar\-(.*)$'),
    ('foo-bar-[фу-бар.*]', '[', ']', r'^foo\-bar\-(фу-бар.*)$'),
    ('a', '{', '}', r'^a$'),
    ('', '{', '}', r'^$'),
    ('[]', '{', '}', r'^\[\]$'),
    ('a[{foo*}]b', '{', '}', r'^a\[(foo*)\]b$'),
    ('[[abc]+]', '[', ']', r'^([abc]+)$'),
    ('foo-[[abc]+]-bar', '[', ']', r'^foo\-([abc]+)\-bar$'),
    ('a-b-{foo-bar.*i{2}}', '{', '}', r'^a\-b\-(foo-bar.*i{2})$'),
])
def test_compile_regex_compiles_correctly(phrase, start, end, output):
    result = compile_regex(phrase, start, end)
    assert output == result.pattern


@pytest.mark.parametrize('phrase, start, end', [
    ('foo:bar:<.*', '<', '>'),
    (']', '[', ']'),
    ('ббб}', '{', '}'),
    ('№', '№', '№'),
    ('[yuy[]', '[', ']'),
    ('yuy[]]', '[', ']'),
    ('[ab[c[[v]]c]', '[', ']'),
])
def test_compile_regex_raises_exception_if_unbalanced(phrase, start, end):
    with pytest.raises(InvalidPatternError) as excinfo:
        compile_regex(phrase, start, end)
    assert phrase in str(excinfo.value)
    assert 'unbalanced braces' in str(excinfo.value)


@pytest.mark.parametrize('phrase, start, end, match_against, should_match_succeed', [
    ('foo:bar:<.*>', '<', '>', 'foo:bar:baz:me:python', True),
    ('foo:bar:<.*>', '<', '>', 'foo:bar:', True),
    ('foo.bar:<.*>', '<', '>', 'foo:bar:', False),
    ('foo:bar:<me*>', '<', '>', 'foo:bar:', False),
    ('foo:bar:<me*>', '<', '>', 'foo:bar:meeeeeeeeee', True),
    ('foo:bar:<me*>', '<', '>', 'foo:bar:m', True),
    ('[[abc]]', '[', ']', 'c', True),
    ('[.*]', '[', ']', '', True),
])
def test_compile_regex_output_matches_correctly(phrase, start, end, match_against, should_match_succeed):
    result = compile_regex(phrase, start, end)
    if should_match_succeed:
        assert result.match(match_against)
    else:
        assert not result.match(match_against)



from vakt.util import JsonSerializer, Subject
from .helper import CountObserver


class AB(JsonSerializer):
    def __init__(self, b):
        self.b = b

    def _data(self):
        return self


class CD(JsonSerializer):
    def _data(self):
        return {'x': 1}


def test_json_serializer_from_json():
    js = AB(5).to_json()
    ab = AB.from_json(js)
    assert isinstance(ab, AB)
    js = CD().to_json()
    cd = CD.from_json(js)
    assert isinstance(cd, dict)
    assert cd == {'x': 1}


def test_observables():
    subj = Subject()
    o1 = CountObserver()
    o2 = CountObserver()
    subj.add_listener(o1)
    subj.add_listener(o2)
    subj.notify()
    subj.notify()
    assert 2 == o1.count
    assert 2 == o2.count
    subj.remove_listener(o1)
    subj.notify()
    assert 2 == o1.count
    assert 3 == o2.count



"""
All Rules relevant to networking context
"""

import ipaddress
import logging
import warnings

from ..rules.base import Rule


log = logging.getLogger(__name__)


__all__ = [
    'CIDR',
]


class CIDR(Rule):
    """
    Rule that is satisfied when inquiry's IP address is in the provided CIDR.
    For example: context={'ip': CIDR('127.0.0.1/32')}
    """

    def __init__(self, cidr):
        self.cidr = cidr

    def satisfied(self, what, inquiry=None):
        if not isinstance(what, str):
            return False
        try:
            ip = ipaddress.ip_address(what)
            net = ipaddress.ip_network(self.cidr)
        except ValueError:
            log.exception('Error %s satisfied', type(self).__name__)
            return False
        return ip in net


# Classes marked for removal in next releases
class CIDRRule(CIDR):
    """Deprecated in favor of CIDR"""
    def __init__(self, *args, **kwargs):
        warnings.warn('CIDRRule will be removed in version 2.0. Use CIDR', DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)



import pytest

from vakt.storage.memory import MemoryStorage
from vakt.policy import Policy
from vakt.guard import Inquiry
from vakt.exceptions import PolicyExistsError
from vakt.rules.operator import Eq
from vakt.rules.logic import Any


@pytest.fixture
def st():
    return MemoryStorage()


def test_add(st):
    st.add(Policy('1', description='foo'))
    assert '1' == st.get('1').uid
    assert 'foo' == st.get('1').description
    st.add(Policy('2', actions=[Eq('get'), Eq('put')], subjects=[Any()], resources=[{'books': Eq('Harry')}]))
    assert '2' == st.get('2').uid
    assert 2 == len(st.get('2').actions)
    assert 1 == len(st.get('2').subjects)
    assert isinstance(st.get('2').subjects[0], Any)
    assert 1 == len(st.get('2').resources)
    assert isinstance(st.get('2').resources[0]['books'], Eq)
    assert 'Harry' == st.get('2').resources[0]['books'].val


def test_policy_create_existing(st):
    st.add(Policy('1', description='foo'))
    with pytest.raises(PolicyExistsError):
        st.add(Policy('1', description='bar'))


def test_get(st):
    st.add(Policy('1'))
    st.add(Policy(2, description='some text'))
    assert isinstance(st.get('1'), Policy)
    assert '1' == st.get('1').uid
    assert 2 == st.get(2).uid
    assert 'some text' == st.get(2).description


@pytest.mark.parametrize('limit, offset, result', [
    (500, 0, 200),
    (101, 1, 101),
    (500, 50, 150),
    (200, 0, 200),
    (200, 1, 199),
    (199, 0, 199),
    (200, 50, 150),
    (0, 0, 0),
    (0, 100, 0),
    (1, 0, 1),
    (5, 4, 5),
    (200, 300, 0),
])
def test_get_all(st, limit, offset, result):
    for i in range(200):
        st.add(Policy(str(i)))
    assert result == len(st.get_all(limit, offset))


def test_get_all_check_policy_properties(st):
    st.add(Policy('1', description='foo'))
    assert 1 == len(st.get_all(100, 0))
    assert '1' == st.get_all(100, 0)[0].uid
    assert 'foo' == st.get_all(100, 0)[0].description


def test_get_all_with_incorrect_args(st):
    with pytest.raises(ValueError) as e:
        st.get_all(-1, 90)
    assert "Limit can't be negative" == str(e.value)

    with pytest.raises(ValueError) as e:
        st.get_all(0, -34)
    assert "Offset can't be negative" == str(e.value)


def test_find_for_inquiry(st):
    st.add(Policy('1', subjects=['max', 'bob']))
    st.add(Policy('2', subjects=['sam', 'nina']))
    st.add(Policy('3', subjects=[Eq('max'), Eq('bob')]))
    inquiry = Inquiry(subject='sam', action='get', resource='books')
    found = st.find_for_inquiry(inquiry)
    found = list(found)
    assert 3 == len(found)
    assert ['max', 'bob'] == found[0].subjects or \
           ['max', 'bob'] == found[1].subjects or \
           ['max', 'bob'] == found[2].subjects


def test_update(st):
    policy = Policy('1')
    st.add(policy)
    assert '1' == st.get('1').uid
    assert None is st.get('1').description
    policy.description = 'foo'
    st.update(policy)
    assert '1' == st.get('1').uid
    assert 'foo' == st.get('1').description
    p = Policy(2, actions=[Any()], subjects=[Eq('max'), Eq('bob')])
    st.add(p)
    assert 2 == st.get(2).uid
    p.actions = [Eq('get')]
    st.update(p)
    assert 1 == len(st.get(2).actions)
    assert 'get' == st.get(2).actions[0].val


def test_delete(st):
    policy = Policy('1')
    st.add(policy)
    assert '1' == st.get('1').uid
    st.delete('1')
    assert None is st.get('1')
    st.delete('1000000')



"""
Base Rule. Should be extended be concrete ones.
"""

import logging
from abc import ABCMeta, abstractmethod

from ..util import JsonSerializer, PrettyPrint
from ..exceptions import RuleCreationError


log = logging.getLogger(__name__)


class Rule(JsonSerializer, PrettyPrint, metaclass=ABCMeta):
    """Basic Rule"""

    @abstractmethod
    def satisfied(self, what, inquiry=None):
        """Is rule satisfied by the inquiry"""
        pass

    @classmethod
    def from_json(cls, data):
        try:
            return cls._parse(data)
        except ValueError as e:
            raise RuleCreationError(str(e))

    def _data(self):
        return self



"""
All Rules for defining Inquiry elements relations
"""
from abc import ABCMeta, abstractmethod

from ..rules.base import Rule


__all__ = [
    'SubjectEqual',
    'ActionEqual',
    'ResourceIn',
    'SubjectMatch',
    'ActionMatch',
    'ResourceMatch',
]


class InquiryMatchAbstract(Rule, metaclass=ABCMeta):
    """
    Base rule for concrete InquiryMatch rule implementations.
    """
    def __init__(self, attribute=None):
        self.attribute = attribute

    def satisfied(self, what, inquiry=None):
        if not inquiry:
            return False
        inquiry_value = getattr(inquiry, self._field_name())
        if self.attribute is not None:
            if isinstance(inquiry_value, dict) and self.attribute in inquiry_value:
                inquiry_value = inquiry_value[self.attribute]
            else:
                return False
        return what == inquiry_value

    @abstractmethod
    def _field_name(self):
        pass


class SubjectMatch(InquiryMatchAbstract):
    """
    Rule that is satisfied if the value equals the Inquiry's Subject or it's attribute.
    For example: resources=[SubjectMatch('nick')]
    """
    def _field_name(self):
        return 'subject'


class ActionMatch(InquiryMatchAbstract):
    """
    Rule that is satisfied if the value equals the Inquiry's Action or it's attribute.
    For example: resources=[{'ref': ActionMatch('ref_method')}]
    """
    def _field_name(self):
        return 'action'


class ResourceMatch(InquiryMatchAbstract):
    """
    Rule that is satisfied if the value equals the Inquiry's Resource or it's attribute.
    For example: resources=[ResourceMatch('sub-category')]
    """
    def _field_name(self):
        return 'resource'


class SubjectEqual(Rule):
    """
    Rule that is satisfied if the string value equals the Inquiry's Subject.
    For example: context={'user': SubjectEqual()}
    This Rule only works for string-based policies.
    """
    def satisfied(self, what, inquiry=None):
        return inquiry and isinstance(what, str) and what == inquiry.subject


class ActionEqual(Rule):
    """
    Rule that is satisfied if the string value equals the Inquiry's Action.
    For example: context={'user': ActionEqual()}
    This Rule only works for string-based policies.
    """
    def satisfied(self, what, inquiry=None):
        return inquiry and isinstance(what, str) and what == inquiry.action


class ResourceIn(Rule):
    """
    Rule that is satisfied if list contains the Inquiry's Resource.
    For example: context={'user': ResourceIn()}
    This Rule only works for string-based policies.
    """
    def satisfied(self, what, inquiry=None):
        return inquiry and isinstance(what, list) and inquiry.resource in what



"""
All Rules that are related to logic and composition:
and, or, not, is-true, is-false.
"""

import logging
from abc import ABCMeta, abstractmethod

from ..rules.base import Rule


log = logging.getLogger(__name__)


__all__ = [
    'Truthy',
    'Falsy',
    'And',
    'Or',
    'Not',
    'Any',
    'Neither',
]


class BooleanRule(Rule, metaclass=ABCMeta):
    """
    Boolean Rule that is satisfied when 'what' is evaluated to a boolean true/false.
    Its `satisfied` accepts:
     - a callable without arguments
     - non-callable
     - expressions
    """
    def satisfied(self, what, inquiry=None):
        res = what() if callable(what) else what
        return bool(res) == self.val

    @property
    @abstractmethod
    def val(self):
        """This should be overridden as a True/False getter property"""
        pass


class Truthy(BooleanRule):
    """
    Rule that is satisfied when 'what' is evaluated to a boolean 'true'.
    For example: subjects=[{'role': Truthy(), 'name': Eq('Jimmy')}]
    """
    @property
    def val(self):
        return True


class Falsy(BooleanRule):
    """
    Rule that is satisfied when 'what' is evaluated to a boolean 'false'.
    For example: subjects=[{'role': Falsy(), 'name': Eq('Jimmy')}]
    """
    @property
    def val(self):
        return False


class CompositionRule(Rule, metaclass=ABCMeta):
    """
    Abstract Rule that encompasses other Rules.
    """
    def __init__(self, *rules):
        for r in rules:
            if not isinstance(r, Rule):
                log.error("%s creation. Arguments should be of Rule class or it's derivatives", type(self).__name__)
                raise TypeError("Arguments should be of Rule class or it's derivatives")
        self.rules = rules


class And(CompositionRule):
    """
    Rule that is satisfied when all the rules it's composed of are satisfied.
    For example: subjects=[{'stars': And(Greater(50), Less(120)), 'name': Eq('Jimmy')}]
    """
    def satisfied(self, what, inquiry=None):
        answers = [x.satisfied(what, inquiry) for x in self.rules]
        return len(answers) > 0 and all(answers)


class Or(CompositionRule):
    """
    Rule that is satisfied when at least one of the rules it's composed of is satisfied.
    Uses short-circuit evaluation.
    For example: subjects=[{'stars': Or(Greater(50), Less(120)), 'name': Eq('Jimmy')}]
    """
    def satisfied(self, what, inquiry=None):
        for rule in self.rules:
            if rule.satisfied(what, inquiry):
                return True
        return False


class Not(Rule):
    """
    Rule that negates another Rule.
    For example: subjects=[{'stars': Eq(555), 'name': Not(Eq('Jimmy'))}]
    """
    def __init__(self, rule):
        if not isinstance(rule, Rule):
            log.error("%s creation. Arguments should be of Rule class or it's derivatives", type(self).__name__)
            raise TypeError("Arguments should be of Rule class or it's derivatives")
        self.rule = rule

    def satisfied(self, what, inquiry=None):
        return not self.rule.satisfied(what, inquiry)


class Any(Rule):
    """
    Rule that is always satisfied.
    For example: resources=[{'endpoint': Any(), 'method': Eq('POST')}]
    """

    def satisfied(self, what=None, inquiry=None):
        return True


class Neither(Rule):
    """
    Rule that always isn't satisfied.
    For example: resources=[{'endpoint': Neither(), 'method': Eq('GET')}]
    """

    def satisfied(self, what=None, inquiry=None):
        return False



import pytest
from operator import attrgetter

from vakt.storage.memory import MemoryStorage
from vakt.policy import Policy
from ..helper import MemoryStorageYieldingExample


@pytest.mark.parametrize('st', [
    MemoryStorage(),
    MemoryStorageYieldingExample(),
])
def test_retrieve_all_for_returning_storage(st):
    st.add(Policy('a'))
    st.add(Policy('b'))
    st.add(Policy('c'))
    st.add(Policy('d'))
    st.add(Policy('e'))
    expected_ids = ['a', 'b', 'c', 'd', 'e']
    for i in range(1, 55):
        result = st.retrieve_all(i)
        result = list(result)
        assert 5 == len(result)
        assert expected_ids == sorted(map(attrgetter('uid'), result))
    # test batch is 0
    assert 0 == len(list(st.retrieve_all(0)))
    # test no batch specified
    res = list(st.retrieve_all())
    assert 5 == len(res)
    assert expected_ids == sorted(map(attrgetter('uid'), res))
    # test too big batch
    res = list(st.retrieve_all(100000))
    assert 5 == len(res)
    assert expected_ids == sorted(map(attrgetter('uid'), res))



import json

from sqlalchemy import Column, Integer, SmallInteger, String, ForeignKey, Text, JSON, Boolean
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from ...policy import Policy, ALLOW_ACCESS, DENY_ACCESS, TYPE_STRING_BASED
from ...rules.base import Rule
from ...parser import compile_regex

Base = declarative_base()


class PolicySubjectModel(Base):
    """Storage model for policy subjects"""

    __tablename__ = 'vakt_policy_subjects'

    id = Column(Integer, primary_key=True)
    uid = Column(String(255), ForeignKey('vakt_policies.uid', ondelete='CASCADE'))
    subject = Column(JSON(), comment='JSON value for rule-based policies')
    subject_string = Column(String(255), index=True, comment='Initial string value for string-based policies')
    subject_regex = Column(String(520),
                           index=True,
                           comment='Regexp from initial string value for string-based policies')


class PolicyResourceModel(Base):
    """Storage model for policy resources"""

    __tablename__ = 'vakt_policy_resources'

    id = Column(Integer, primary_key=True)
    uid = Column(String(255), ForeignKey('vakt_policies.uid', ondelete='CASCADE'))
    resource = Column(JSON(), comment='JSON value for rule-based policies')
    resource_string = Column(String(255), index=True, comment='Initial string value for string-based policies')
    resource_regex = Column(String(520),
                            index=True,
                            comment='Regexp from initial string value for string-based policies')


class PolicyActionModel(Base):
    """Storage model for policy actions"""

    __tablename__ = 'vakt_policy_actions'

    id = Column(Integer, primary_key=True)
    uid = Column(String(255), ForeignKey('vakt_policies.uid', ondelete='CASCADE'))
    action = Column(JSON(), comment='JSON value for rule-based policies')
    action_string = Column(String(255), index=True, comment='Initial string value for string-based policies')
    action_regex = Column(String(520),
                          index=True,
                          comment='Regexp from initial string value for string-based policies')


class PolicyModel(Base):
    """Storage model for policy"""

    __tablename__ = 'vakt_policies'

    uid = Column(String(255), primary_key=True)
    type = Column(SmallInteger)
    description = Column(Text())
    effect = Column(Boolean())
    context = Column(JSON())
    subjects = relationship(PolicySubjectModel, passive_deletes=True, lazy='joined')
    resources = relationship(PolicyResourceModel, passive_deletes=True, lazy='joined')
    actions = relationship(PolicyActionModel, passive_deletes=True, lazy='joined')

    @classmethod
    def from_policy(cls, policy):
        """
            Instantiate from policy object

            :param policy: object of type policy
        """
        rvalue = cls()
        return cls._save(policy, model=rvalue)

    def update(self, policy):
        """
            Update object attributes to match given policy

            :param policy: object of type policy
        """
        self._save(policy, model=self)

    def to_policy(self):
        """
            Create a policy object

            :return: object of type `Policy`
        """
        return Policy(uid=self.uid,
                      effect=ALLOW_ACCESS if self.effect else DENY_ACCESS,
                      description=self.description,
                      context=Rule.from_json(self.context),
                      subjects=[
                          self._policy_element_from_db(self.type, x.subject, x.subject_string)
                          for x in self.subjects
                      ],
                      resources=[
                          self._policy_element_from_db(self.type, x.resource, x.resource_string)
                          for x in self.resources
                      ],
                      actions=[
                          self._policy_element_from_db(self.type, x.action, x.action_string)
                          for x in self.actions
                      ])

    @classmethod
    def _save(cls, policy, model):
        """
            Helper to create PolicyModel from Policy object for add and update operations.

            :param policy: object of type Policy
            :param model: object of type PolicyModel
        """
        policy_json = policy.to_json()
        policy_dict = json.loads(policy_json)
        model.uid = policy_dict['uid']
        model.type = policy_dict['type']
        model.effect = policy_dict['effect'] == ALLOW_ACCESS
        model.description = policy_dict['description']
        model.context = json.dumps(policy_dict['context'])
        model.subjects = [
            PolicySubjectModel(subject=x, subject_string=string, subject_regex=compiled)
            for y in policy_dict['subjects']
            for (x, string, compiled) in cls._policy_element_to_db(policy, y)
        ]
        model.resources = [
            PolicyResourceModel(resource=x, resource_string=string, resource_regex=compiled)
            for y in policy_dict['resources']
            for (x, string, compiled) in cls._policy_element_to_db(policy, y)
        ]
        model.actions = [
            PolicyActionModel(action=x, action_string=string, action_regex=compiled)
            for y in policy_dict['actions']
            for (x, string, compiled) in cls._policy_element_to_db(policy, y)
        ]
        return model

    @classmethod
    def _policy_element_to_db(cls, policy, el):
        json_value, string_value, compiled = None, None, None
        if policy.type == TYPE_STRING_BASED:
            string_value = el
            if policy.start_tag in el and policy.end_tag in el:
                compiled = compile_regex(el, policy.start_tag, policy.end_tag).pattern
        else:  # it's a rule-based policy and it's value is a json
            json_value = json.dumps(el)
        yield (json_value, string_value, compiled)

    @classmethod
    def _policy_element_from_db(cls, policy_type, element_json, element_string):
        if policy_type == TYPE_STRING_BASED:
            return element_string
        return Rule.from_json(element_json)



import pytest

from vakt.storage.memory import MemoryStorage
from vakt.storage.observable import ObservableMutationStorage
from ..helper import CountObserver
from vakt import Policy, Inquiry


class TestObservableMutationStorage:

    @pytest.fixture()
    def factory(self):
        def objects_factory():
            mem = MemoryStorage()
            st = ObservableMutationStorage(mem)
            observer = CountObserver()
            st.add_listener(observer)
            return st, mem, observer
        return objects_factory

    def test_add(self, factory):
        st, mem, observer = factory()
        p1 = Policy(1)
        p2 = Policy(2)
        p3 = Policy(3)
        st.add(p1)
        st.add(p2)
        assert 2 == observer.count
        assert 2 == len(list(mem.retrieve_all()))
        assert 2 == len(list(st.retrieve_all()))
        st.add(p3)
        assert 3 == observer.count

    def test_get(self, factory):
        st, mem, observer = factory()
        p1 = Policy('a')
        st.add(p1)
        assert 'a' == st.get('a').uid
        assert 'a' == mem.get('a').uid
        assert 1 == observer.count
        assert None is st.get('b')
        assert None is mem.get('b')
        assert 1 == observer.count

    def test_update(self, factory):
        st, mem, observer = factory()
        p1 = Policy('a')
        st.add(p1)
        p1.uid = 'b'
        st.update(p1)
        assert 'b' == list(mem.retrieve_all())[0].uid
        assert 'b' == list(st.retrieve_all())[0].uid
        assert 2 == observer.count
        p1.uid = 'c'
        st.update(p1)
        assert 3 == observer.count

    def test_delete(self, factory):
        st, mem, observer = factory()
        p1 = Policy('a')
        st.add(p1)
        st.delete('a')
        assert [] == list(mem.retrieve_all())
        assert [] == list(st.retrieve_all())
        assert 2 == observer.count
        st.delete('a')
        assert [] == list(mem.retrieve_all())
        assert [] == list(st.retrieve_all())
        assert 3 == observer.count

    def test_retrieve_all(self, factory):
        st, mem, observer = factory()
        p1 = Policy('a')
        st.add(p1)
        assert 'a' == list(mem.retrieve_all())[0].uid
        assert 'a' == list(st.retrieve_all())[0].uid
        assert 1 == observer.count
        assert 'a' == list(mem.retrieve_all(batch=2))[0].uid
        assert 'a' == list(st.retrieve_all(batch=2))[0].uid
        assert 1 == observer.count
        assert 'a' == list(mem.retrieve_all(5))[0].uid
        assert 'a' == list(st.retrieve_all(5))[0].uid
        assert 1 == observer.count

    def test_get_all(self, factory):
        st, mem, observer = factory()
        p1 = Policy('a')
        st.add(p1)
        assert 'a' == list(mem.get_all(5, 0))[0].uid
        assert 'a' == list(st.get_all(5, 0))[0].uid
        assert 1 == observer.count
        st.get_all(9, 0)
        st.get_all(888, 0)
        assert 1 == observer.count

    def test_find_for_inquiry(self, factory):
        st, mem, observer = factory()
        inq = Inquiry(action='get', subject='foo', resource='bar')
        p1 = Policy('a')
        p2 = Policy('b')
        st.add(p1)
        st.add(p2)
        mem_found = list(mem.find_for_inquiry(inq))
        assert [p1, p2] == mem_found or [p2, p1] == mem_found
        st_found = list(st.find_for_inquiry(inq))
        assert [p1, p2] == st_found or [p2, p1] == st_found
        assert 2 == observer.count
        st.find_for_inquiry(inq)
        st.find_for_inquiry(inq)
        assert 2 == observer.count



import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session

from vakt.storage.sql import SQLStorage
from vakt.storage.sql.migrations import SQLMigrationSet, Migration0To1x3x0
from vakt.storage.sql.model import Base, PolicyActionModel, PolicySubjectModel, PolicyResourceModel, PolicyModel

from . import create_test_sql_engine


@pytest.mark.sql_integration
class TestSQLMigrationSet:

    @pytest.yield_fixture
    def session(self):
        engine = create_test_sql_engine()
        session = scoped_session(sessionmaker(bind=engine))
        yield session
        Base.metadata.drop_all(engine)

    @pytest.yield_fixture
    def migration_set(self, session):
        storage = SQLStorage(scoped_session=session)
        yield SQLMigrationSet(storage)
        session.remove()

    def test_application_of_migration_number(self, migration_set):
        assert 0 == migration_set.last_applied()
        migration_set.save_applied_number(6)
        assert 6 == migration_set.last_applied()
        migration_set.save_applied_number(2)
        assert 2 == migration_set.last_applied()

    def test_up_and_down(self, migration_set):
        migration_set.save_applied_number(0)
        migration_set.up()
        assert 1 == migration_set.last_applied()
        migration_set.up()
        assert 1 == migration_set.last_applied()
        migration_set.down()
        assert 0 == migration_set.last_applied()
        migration_set.down()
        assert 0 == migration_set.last_applied()


@pytest.mark.sql_integration
class TestMigration0To1x3x0:

    @pytest.yield_fixture
    def engine(self):
        engine = create_engine('sqlite:///:memory:', echo=False)
        yield engine

    @pytest.yield_fixture
    def session(self, engine):
        session = scoped_session(sessionmaker(bind=engine))
        yield session
        Base.metadata.drop_all(engine)

    @pytest.yield_fixture
    def migration(self, session):
        storage = SQLStorage(scoped_session=session)
        yield Migration0To1x3x0(storage)
        session.remove()

    def test_order(self, migration):
        assert 1 == migration.order

    def test_has_access_to_storage(self, migration):
        assert hasattr(migration, 'storage') and migration.storage is not None

    def test_up(self, migration, engine):
        migration.up()
        assert Base.metadata.tables[PolicyModel.__tablename__].exists(engine)
        assert Base.metadata.tables[PolicySubjectModel.__tablename__].exists(engine)
        assert Base.metadata.tables[PolicyResourceModel.__tablename__].exists(engine)
        assert Base.metadata.tables[PolicyActionModel.__tablename__].exists(engine)

    def test_down(self, migration, engine):
        migration.down()
        assert not Base.metadata.tables[PolicyModel.__tablename__].exists(engine)
        assert not Base.metadata.tables[PolicySubjectModel.__tablename__].exists(engine)
        assert not Base.metadata.tables[PolicyResourceModel.__tablename__].exists(engine)
        assert not Base.metadata.tables[PolicyActionModel.__tablename__].exists(engine)



from setuptools import setup, find_packages
from os import path


here = path.abspath(path.dirname(__file__))
about = {}
with open(path.join(here, 'vakt', 'version.py'), mode='r', encoding='utf-8') as f:
    exec(f.read(), about)

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


if __name__ == '__main__':
    setup(
        name='vakt',
        description='Attribute-based access control (ABAC) SDK for Python',
        keywords='ACL ABAC access-control policy security authorization permission',
        version=about['__version__'],
        author='Egor Kolotaev',
        author_email='ekolotaev@gmail.com',
        license="Apache 2.0 license",
        url='http://github.com/kolotaev/vakt',
        long_description=long_description,
        long_description_content_type='text/markdown',
        py_modules=['vakt'],
        python_requires='>=3.4',
        install_requires=[
            'jsonpickle~=1.0',
        ],
        extras_require={
            'dev': [
                'pytest~=4.0',
                'pytest-cov~=2.6',
                'pylint~=1.0',
                'PyMySQL~=0.9',
                'psycopg2cffi~=2.7',
            ],
            'mongo': [
                'pymongo~=3.5',
            ],
            'sql': [
                'SQLAlchemy~=1.3'
            ],
        },
        packages=find_packages(exclude='tests'),
        classifiers=[
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
            'Topic :: System :: Systems Administration',
            'Topic :: System :: Networking',
            'Topic :: System :: Networking :: Firewalls',
            'Topic :: Security',
            'Topic :: Software Development',
            'Topic :: Utilities',
            'Natural Language :: English',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: Implementation :: PyPy',
            'Programming Language :: Python :: Implementation :: CPython',
        ],
    )



"""
Namespace for a basic Policy class.
"""

import logging
import warnings
import copy

from .effects import ALLOW_ACCESS, DENY_ACCESS
from .exceptions import PolicyCreationError
from .util import JsonSerializer, PrettyPrint
from .rules.base import Rule


log = logging.getLogger(__name__)


# Types for Policies and Inquiries:
# String-based (simple strings, regexps)
TYPE_STRING_BASED = 1
# Rule-based definitions (Rules).
TYPE_RULE_BASED = 2


class Policy(JsonSerializer, PrettyPrint):
    """Represents a policy that regulates access and allowed actions of subjects
    over some resources under a set of context restrictions.
    """

    # Fields that affect Policy definition and further logic for `fit`.
    _definition_fields = ['subjects', 'resources', 'actions']

    def __init__(self, uid, subjects=(), effect=DENY_ACCESS, resources=(),
                 actions=(), context=None, rules=None, description=None):
        self.uid = uid
        self.subjects = subjects
        self.effect = effect or DENY_ACCESS
        self.resources = resources
        self.actions = actions
        # check for deprecated rules argument.
        # If both 'context' and 'rules' are present - 'context' wins
        if context is not None:
            pass
        elif rules:
            warnings.warn("'rules' argument will be removed in version 2.0. Use 'context' argument",
                          DeprecationWarning, stacklevel=2)
            context = rules
        else:
            context = {}
        self.context = context
        self.description = description
        self.type = None

    @classmethod
    def from_json(cls, data):
        props = cls._parse(data)
        if 'uid' not in props:
            log.error("Error creating policy from json. 'uid' attribute is required")
            raise PolicyCreationError("Error creating policy from json. 'uid' attribute is required")
        context_rules = {}
        if 'context' in props:
            context_rules = props['context']
        elif 'rules' in props:  # this is to support deprecated 'rules' attribute
            context_rules = props['rules']
            del props['rules']
        props['context'] = context_rules
        if 'type' in props:  # type is calculated dynamically on init
            del props['type']
        return cls(**props)

    def allow_access(self):
        """Does policy imply allow-access?"""
        return self.effect == ALLOW_ACCESS

    @property
    def start_tag(self):
        """
        Policy expression start tag.
        Used for policies defined with regexp.
        """
        return '<'

    @property
    def end_tag(self):
        """
        Policy expression end tag.
        Used for policies defined with regexp.
        """
        return '>'

    def __setattr__(self, name, value):
        self._check_field_type(name, value)
        # Always calculate type. Even if type is set explicitly.
        calculated_type = self._calculate_type(name, value)
        object.__setattr__(self, name, value)
        # Dict assign eliminates recursion
        self.__dict__['type'] = calculated_type

    def _calculate_type(self, new_element_name, new_element_value):
        all_elements = rule_elements = str_elements = 0
        self_copy = copy.copy(self)
        self_copy.__dict__[new_element_name] = new_element_value
        for elements in [getattr(self_copy, f, ()) for f in self._definition_fields]:
            for e in elements:
                all_elements += 1
                if isinstance(e, (dict, Rule)):
                    rule_elements += 1
                elif isinstance(e, str):
                    str_elements += 1
        if all_elements == str_elements or all_elements == 0:
            return TYPE_STRING_BASED
        if all_elements == rule_elements:
            return TYPE_RULE_BASED
        raise PolicyCreationError(
            'Policy elements should all be either dict, Rule (for rule-based) or string (for string-based)'
        )

    def _check_field_type(self, name, value):
        """Checks type of a field that defines Policy"""
        if name in self._definition_fields and not all(map(lambda x: isinstance(x, (str, dict, Rule)), value)):
            raise PolicyCreationError(
                'Field "%s" element must be of `str`, `dict` or `Rule` type. But given: %s' % (name, value)
            )
        if name == 'context' and not isinstance(value, dict):
            raise PolicyCreationError('Error creating Policy. Context must be a dictionary')

    def _data(self):
        data = vars(self)
        # get rid of "py/tuple" for tuples upon serialization in favor of simple json-array
        for k, prop in data.items():
            if isinstance(prop, tuple):
                data[k] = list(prop)
        return data


class PolicyAllow(Policy):
    """
    Policy that has effect ALLOW_ACCESS by default.
    """
    def __init__(self, uid, subjects=(), resources=(), actions=(), context=None, description=None):
        super().__init__(uid, effect=ALLOW_ACCESS,
                         subjects=subjects, resources=resources,
                         actions=actions, context=context, description=description)


class PolicyDeny(Policy):
    """
    Policy that has effect DENY_ACCESS by default.
    """
    def __init__(self, uid, subjects=(), resources=(), actions=(), context=None, description=None):
        super().__init__(uid, effect=DENY_ACCESS,
                         subjects=subjects, resources=resources,
                         actions=actions, context=context, description=description)



from ..util import Subject


class ObservableMutationStorage(Subject):
    """
    Wraps Storage.
    Implements mutation part of Storage interface as a notifier of subscribers.
    Notifies observers when mutation method is called on Storage.
    Read part of Storage interface is a simple proxy.
    """
    def __init__(self, storage):
        self.storage = storage
        super().__init__()

    def add(self, policy):
        res = self.storage.add(policy)
        self.notify()
        return res

    def update(self, policy):
        res = self.storage.update(policy)
        self.notify()
        return res

    def delete(self, uid):
        res = self.storage.delete(uid)
        self.notify()
        return res

    def get(self, uid):
        return self.storage.get(uid)

    def get_all(self, limit, offset):
        return self.storage.get_all(limit, offset)

    def retrieve_all(self, *args, **kwargs):
        return self.storage.retrieve_all(*args, **kwargs)

    def find_for_inquiry(self, inquiry, checker=None):
        return self.storage.find_for_inquiry(inquiry, checker)
